# coding: utf-8

import numpy as np
from numpy.linalg import norm
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec

# Model Parameter
V0 = 16.
beta = 1.
d = 2
shift = 1.1

def pot(xin, V0=V0, beta=beta, shift=shift):
    x = xin-shift
    return V0+V0*(np.exp(-2*beta*x) - 2*np.exp(-beta*x) )

def Dpot(xin, V0=V0, beta=beta, shift=shift):
    x = xin-shift
    return -2*beta*V0*(np.exp(-2*beta*x) - np.exp(-beta*x) )

def potv(x,y): # Potential 2D Form wurde zum Plotten gebraucht
    no = np.sqrt(x**2 + y**2)
    return pot(no)

# Anfangsdaten

q0 = 0.5*np.ones(d)
p0 = np.zeros(d)
p0[0] = 1.
y0 = np.array((q0,p0)).ravel()

def Ekin(y):
    p = y[d:2*d]
    res = 0.5*np.diag(p.T @ p)
    return res

def Epot(y):
    q = y[:d]
    nq = np.sqrt( np.diag(q.T @ q) )
    return pot(nq)

def Energy(y):
    return Ekin(y) + Epot(y)

def compute_energies(y):
    EK = Ekin(y)
    EP = Epot(y)
    ET = EK+EP
    return EK, EP, ET

def plot_solutions(t, y, ekin, epot, etot, data='', filemark=''):
    fig = plt.figure()
    fig.suptitle(data)
    gs = GridSpec(2, 2, figure=fig)
    ax0 = fig.add_subplot(gs[0, :])
    
    ax0.plot(t, ekin, label='ekin ' + data)
    ax0.plot(t, epot, label='epot ' + data)
    ax0.plot(t, etot, label='etot ' + data)
    ax0.set_ylabel('Energy')

    ax0r = ax0.twinx()  # instantiate a second axes that shares the same x-axis
    colorsteps = 'tab:red'
    ax0r.plot(t[:-1], np.diff(t),'.' ,label='stepsize ' + data, color=colorsteps)
    ax0r.set_ylabel('Stepsize', color=colorsteps)
    ax0r.tick_params(axis='y', labelcolor=colorsteps)
    ax0.legend()

    # plot potential and trajectory in space
    ax1 = fig.add_subplot(gs[1, 0])

    # prepare ploting the potential
    q1min = y[0].min()
    q1max = y[0].max()
    q2min = y[1].min()
    q2max = y[1].max()
    #
    x1 = np.linspace(q1min,q1max,100)
    x2 = np.linspace(q2min,q2max,100)
    X, Y = np.meshgrid(x1, x2)
    #
    Z = potv(X,Y)
    # plot the potential and a few contour lines
    levels = np.array([0.5,1.,1.5,2.,3.,4.,5.,6.,8.,12]) # backed for this potential
    contours = ax1.contour(X,Y,Z,levels=levels, colors='black', linewidths=0.5)
    ax1.clabel(contours, inline=True, fontsize=6)
    im1  = ax1.imshow(Z, extent = [q1min, q1max, q2min, q2max], origin = 'lower', cmap='RdGy', alpha=0.5)
    clb = fig.colorbar(im1, ax=ax1)
    clb.ax.get_yaxis().labelpad = 15
    clb.set_label('Potential', rotation= 270)
    
    # plot the trajectory in space
    ax1.plot(y[0], y[1], '.-',  label = 'q')
    ax1.legend()

    # plot the trajectory in momentum
    ax2 = fig.add_subplot(gs[1, 1])
    ax2.plot(y[2], y[3],  '.-', label = 'p')
    ax2.axis('equal')
    ax2.legend()
    
    fig.tight_layout()  # otherwise the right y-label is slightly clipped
    filename = data + filemark 
    plt.savefig(filename+'01.pdf')
    plt.savefig(filename+'01.png')
    plt.show()


# (a) auf Papier
# (b) auf Papier
# (c) und (d)

# -------- Schritte im Splitting ------------
# u enthaelt
#              q in den ersten d Komponenten und
#              p in den nexten d Komponenten und
#              t am Ende

def phif(dt, u):
    v = np.zeros_like(u)
    # TODO: Implementierung
    return v

def phig(dt, u):
    v = np.zeros_like(u)
    # TODO: Implementierung
    return v


# ------------- Rechte Seite der Differentialgleichung -----
# t muss vor y als Variable erscheinen 
# y enthaellt
#              q in den ersten d Komponenten und
#              p in den nexten d Komponenten 
def rhs(t, y):
    v = np.zeros_like(y)
    # TODO: Implementierung
    return v


print("-------------- Kurzzeit Integration -----------------")
tstart = 0.
Tend = 16.75
tspan = (tstart, Tend)
N = 400 # Anzahl Schritte im Splitting

# Anfangswert fuer Splitting
u0 = np.zeros(2*d+1)
u0[:d] = q0
u0[d:-1] = p0
u0[-1] = tstart
ys = 1.* u0 # Backup Anfangswert

method = 'BM42'

# TODO: Ersetzen Sie die folgenden 2 Zeilen mit der Implementierung zur Kurrzeit-Integration mit BM42 
tsplit = np.zeros(N+1)
ysplit = np.zeros((N+1,2*d+1))


ysplit = ysplit.T
print('intsplit '+method,':\n', tsplit[-1], ysplit[:,-1], tsplit.shape)
ek, ep, et = compute_energies(ysplit)
plot_solutions(tsplit, ysplit, ek, ep, et, data=method, filemark='kurz')

ys = 1.* u0 # Backup Anfangswert

method = 'BM63'

# TODO: Ersetzen Sie die folgenden 2 Zeilen mit der Implementierung zur Kurrzeit-Integration mit BM63 
tsplit = np.ones(N+1)
ysplit = np.ones((N+1,2*d+1))

ysplit = ysplit.T
print('intsplit '+method,':\n', tsplit[-1], ysplit[:,-1], tsplit.shape)
ek, ep, et = compute_energies(ysplit)
plot_solutions(tsplit, ysplit, ek, ep, et, data=method, filemark='kurz')

# TODO: Ersetzen Sie die folgenden 2 Zeilen mit der Implementierung zur Kurrzeit-Integration mit standard integrator
tstand = tsplit
ystand = ysplit

print('solve_ivp:',tstand[-1], ystand[:,-1], tstand.shape)
ek, ep, et = compute_energies(ystand)
plot_solutions(tstand, ystand, ek, ep, et, data='ivp', filemark='kurz')

# (e) 
print('-------------  Langzeit Integration -----------------------')

Tend = 800.
tspan = (tstart, Tend)

# TODO: Ersetzen Sie die folgenden 2 Zeilen mit der Implementierung zur Kurrzeit-Integration mit standard integrator
tstand = tsplit
ystand = ysplit

# Nachbearbeitung und Darstellung der Ergebnisse
print('solve_ivp:',tstand[-1], ystand[:,-1], tstand.shape)
ek_ivp, ep_ivp, et_ivp = compute_energies(ystand)

edev_ivp = et_ivp[1:]-et_ivp[0]
dts = np.diff(tstand)
print('IVP: Max Deviation in Energy: ',abs(edev_ivp).max(), ' with min stepsize', dts.min())

method = 'BM42'
N = 5000*3 # BM42
ys = 1.* u0 # Backup Anfangswert

# TODO: Ersetzen Sie die folgenden 2 Zeilen mit der Implementierung zur Kurrzeit-Integration mit BM42
tsplit = np.ones(N+1)
ysplit = np.ones((N+1,2*d+1))

# Nachbearbeitung und Darstellung der Ergebnisse
ysplit = ysplit.T
print('intsplit '+method,':\n', tsplit[-1], ysplit[:,-1], tsplit.shape)

ek_s, ep_s, et_s = compute_energies(ysplit)
edev_s = et_s[1:]-et_s[0]
print(method+': Max Deviation in Energy: ',abs(edev_s).max(), ' with stepsize', (Tend-tstart)/N)

plt.figure()
plt.semilogy(tsplit[1:], abs(edev_s), label=method)
plt.semilogy(tstand[1:], abs(edev_ivp), label='ivp')
plt.legend()
plt.savefig('ivp_'+method+'.pdf')
plt.savefig('ivp_'+method+'.png')
plt.show()


method = 'BM63'
N = 5000*2 # BM63
ys = 1.* u0  # Backup Anfangswert

# TODO: Ersetzen Sie die folgenden 2 Zeilen mit der Implementierung zur Kurrzeit-Integration mit BM63 
tsplit = np.ones(N+1)
ysplit = np.ones((N+1,2*d+1))

# Nachbearbeitung und Darstellung der Ergebnisse
ysplit = ysplit.T
print('intsplit '+method,':\n', tsplit[-1], ysplit[:,-1], tsplit.shape)

ek_s, ep_s, et_s = compute_energies(ysplit)
edev_s = et_s[1:]-et_s[0]
print(method+': Max Deviation in Energy: ',abs(edev_s).max(), ' with stepsize', (Tend-tstart)/N)

plt.figure()
plt.semilogy(tsplit[1:], abs(edev_s), label=method)
plt.semilogy(tstand[1:], abs(edev_ivp), label='ivp')
plt.legend()
plt.savefig('ivp_'+method+'.pdf')
plt.savefig('ivp_'+method+'.png')
plt.show()

# (e) weiter auf Papier 
